#include "ZUNO_DHT.h"
#include "EEPROM.h"

#define TEMPERATURE_CHANNEL 1
#define HUMIDITY_CHANNEL 2
ZUNO_SETUP_CHANNELS(
  ZUNO_SENSOR_MULTILEVEL(
    ZUNO_SENSOR_MULTILEVEL_TYPE_TEMPERATURE, 
    SENSOR_MULTILEVEL_SCALE_CELSIUS, 
    SENSOR_MULTILEVEL_SIZE_TWO_BYTES, 
    SENSOR_MULTILEVEL_PRECISION_ONE_DECIMAL, 
    getTemperature
  ),
  ZUNO_SENSOR_MULTILEVEL(
    ZUNO_SENSOR_MULTILEVEL_TYPE_RELATIVE_HUMIDITY, 
    SENSOR_MULTILEVEL_SCALE_PERCENTAGE_VALUE, 
    SENSOR_MULTILEVEL_SIZE_TWO_BYTES, 
    SENSOR_MULTILEVEL_PRECISION_ONE_DECIMAL, 
    getHumidity
  )
);

ZUNO_SETUP_ASSOCIATIONS(
  ZUNO_ASSOCIATION_GROUP_SET_VALUE
);

ZUNO_SETUP_SLEEPING_MODE(ZUNO_SLEEPING_MODE_FREQUENTLY_AWAKE);

ZUNO_SETUP_CFGPARAMETER_HANDLER(configHandler);

ZUNO_SETUP_BATTERY_HANDLER(batteryHandler);

#define ZWAVE_CONFIG_DEFAULT 96
#define ZWAVE_CONFIG_WAKEUP_INTERVAL 64 // wakeup interval in seconds
#define ZWAVE_CONFIG_TEMPERATURE_RESOLUTION 65 // temperature resolution in 1/10 of celsius degre, only report when value change
#define ZWAVE_CONFIG_HUMIDITY_RESOLUTION 66 // temperature resolution in 1/10 of percent unit, only report when value change
struct ConfigStruct {
  uint16_t wakeupInterval;
  float temperatureResolution;
  float humidityResolution;
};
ConfigStruct config = {10, 0.5, 1.0};

struct ReadingsStruct {
  float temperature;
  float humidity;
};
ReadingsStruct readings;

DHT dht22(11, DHT22);
byte n;

#define MAGIC_BYTE 42
#define CAN_PRECISION_BITS 8
#define VCC_VOLTAGE 3.3 // 3.1 from internal regulator, 3.3 from booster
#define MAX_BATTERY_VOLTAGE 3.0

#define LED_PIN LED_BUILTIN
#define BAT_PIN A3 // A3 ADC3 11
#define DEBUG

#ifdef DEBUG
  #define debug(...) {__VA_ARGS__}
#else
  #define debug(...)
#endif

void setup() {
  // Setup Z-Uno pins
  pinMode(LED_PIN, OUTPUT);
  digitalWrite(LED_PIN, LOW);

  pinMode(BAT_PIN, INPUT); 
  analogReadResolution(CAN_PRECISION_BITS);
  analogReference(ADC_HIGH_VCC | ADC_LOW_GND);

  // Setup logging
  debug(Serial.begin(115200);)

  // Load configuration
  if (zunoLoadCFGParam(ZWAVE_CONFIG_DEFAULT) == MAGIC_BYTE) {
    config.wakeupInterval = zunoLoadCFGParam(ZWAVE_CONFIG_WAKEUP_INTERVAL);
    config.temperatureResolution = zunoLoadCFGParam(ZWAVE_CONFIG_TEMPERATURE_RESOLUTION) / 10.0;
    config.humidityResolution = zunoLoadCFGParam(ZWAVE_CONFIG_HUMIDITY_RESOLUTION) / 10.0;
  } else {
    zunoSaveCFGParam(ZWAVE_CONFIG_WAKEUP_INTERVAL, config.wakeupInterval);
    zunoSaveCFGParam(ZWAVE_CONFIG_TEMPERATURE_RESOLUTION, config.temperatureResolution * 10);
    zunoSaveCFGParam(ZWAVE_CONFIG_HUMIDITY_RESOLUTION, config.humidityResolution * 10);
    zunoSaveCFGParam(ZWAVE_CONFIG_DEFAULT, MAGIC_BYTE);
  }
  
  // Initialize DHT22 library
  dht22.begin();
}

float roundToResolution(float value, float resolution) {
  return round(value / resolution) * resolution;
}

void loadReadings() {
  NZRAM.get(0x0, &n, 1);
  if (n == MAGIC_BYTE) {
    debug(Serial.println("Found previous readings, loading...");)
    NZRAM.get(0x1, &readings, sizeof(readings));
  }
}

void saveReadings() {
  NZRAM.put(0x1, &readings, sizeof(readings));
  n = MAGIC_BYTE;
  NZRAM.put(0x0, &n, 1);
}

void loop() {
  debug(Serial.print("Sketch is starting... (wake up reason: "); Serial.print(zunoGetWakeReason(), DEC); Serial.println(")");)
  
  digitalWrite(LED_PIN, HIGH);
  
  dht22.read(true);
  float temperature = roundToResolution(dht22.readTemperature(), config.temperatureResolution);
  float humidity = roundToResolution(dht22.readHumidity(), config.humidityResolution);

  debug(Serial.println("---"); Serial.print("Millis: "); Serial.print(millis()); Serial.println("ms");)

  loadReadings();
  
  // Compare and publish temperature
  debug(Serial.print("Temperature: "); Serial.print(temperature); Serial.println("°C");)
  if (temperature != readings.temperature) {
    readings.temperature = temperature;
    debug(Serial.println(" -> reporting value");)
    zunoSendReport(TEMPERATURE_CHANNEL);
    zunoSendToGroupSetValueCommand(CTRL_GROUP_1, temperature);  
  }

  // Compare and publish humidity
  debug(Serial.print("Humidity: "); Serial.print(humidity); Serial.println("%");)
  if (humidity != readings.humidity) {
    readings.humidity = humidity;
    debug(Serial.println(" -> reporting value");)
    zunoSendReport(HUMIDITY_CHANNEL);
  }

  saveReadings();

  digitalWrite(LED_PIN, LOW);
  
  zunoSetBeamCountWU(config.wakeupInterval);
  zunoSendDeviceToSleep();
}

word getTemperature(void) {
  loadReadings();
  return readings.temperature * 10;
}

word getHumidity(void) {
  loadReadings();
  return readings.humidity * 10;
}

void configHandler(byte param, word value) {
  switch(param) {
    case ZWAVE_CONFIG_WAKEUP_INTERVAL:
      config.wakeupInterval = value;
      break;
    case ZWAVE_CONFIG_TEMPERATURE_RESOLUTION:
      config.temperatureResolution = value / 10.0;
      break;
    case ZWAVE_CONFIG_HUMIDITY_RESOLUTION:
      config.humidityResolution = value / 10.0;
      break;
  }
}

byte batteryHandler() {
  // Constants
  const unsigned int CAN_RESOLUTION = pow(2, CAN_PRECISION_BITS);
  const unsigned int MAX_VALUE = CAN_RESOLUTION * MAX_BATTERY_VOLTAGE / VCC_VOLTAGE;
  // Measurement
  unsigned int value = 0;
  for (byte i = 0; i < 10; i++)
    value += analogRead(BAT_PIN);
  value /= 10;
  // Calculation
  float voltage = (float)value * VCC_VOLTAGE / CAN_RESOLUTION;
  byte percentage = constrain(map(value, 0, MAX_VALUE, 0, 100), 0, 100);
  debug(Serial.print("Battery: "); Serial.print(voltage); Serial.print("V"); Serial.print(" ("); Serial.print(percentage); Serial.println("%)");)
  return percentage;
}
